'use strict';
/*
	TestCase
	Abstract class implementing minimal JUnit 5 protocol
	*/

//----------------------------------------------------	

//==============================================

class TestCase {
	/*ivars:
	Note: don't start ivars with 'test' as that is how test selectors are named
	subTests (lazily initialized)	
	count;
	failureCount;
	exceptionCount;
	currentSubTest;

	log*/
	
	constructor() {
		this.clear();
	}
	
	hasSubTestWithFailure() {
		for (let eachTest of this.getSubTests())
			if (eachTest.hasFailureOrExeption())
				return true;
		return false;
	}
	
	isPassing() {
		return !this.hasSubTestWithFailure();
	}
	
	statusLabel() {
		return this.isPassing() ? 'PASSED' : 'ERRORS';
	}
	
	setCurrentSubTest(subTest) {
		this.currentSubTest = subTest;
	}
	
	assertEquals(expected, actual, optionalMsg) {
		this.basicAssertEquals(expected, actual, true, optionalMsg);
	}	
	
	assertNotEquals(expected, actual, optionalMsg) {
		this.currentSubTest.incrementCount();
		//this.count++;
		if (this.assertEqualsSafely(expected, actual)) {
			//this.failureCount++;
			this.currentSubTest.incrementFailures();
			this.show(`FAILED (assertNotEquals) -- actual <${actual}> should be NOT equal to expected <${expected}> ${this.conformAssertMsg(optionalMsg)}\n`); 			
		}
	}	
	
	assertTrue(condition, optionalMsg) {
		this.basicAssertEquals(true, condition, optionalMsg === undefined, optionalMsg);
		
	}
	
	basicAssertEquals(expected, actual, shouldCompare, optionalMsg) {
		this.currentSubTest.incrementCount();	
		//this.count++;
		if (!this.assertEqualsSafely(expected, actual)) {
			//this.failureCount++;
			this.currentSubTest.incrementFailures();
			this.show(`FAILED (${this.currentSubTest.getTestLabel()})${this.conformCompare(expected, actual, shouldCompare)}${this.conformAssertMsg(optionalMsg)}`); 
		}
	}
	
	assertFloatEquals(expected, actual, tolerance, optionalMsg) {
		//this.count++;	
		this.currentSubTest.incrementCount();			
		if (Math.abs(actual - expected) > Math.abs(tolerance)) {
			//this.failureCount++;
			this.currentSubTest.incrementFailures();
			this.show(`FAILED (${this.currentSubTest.getTestLabel()})${this.conformCompare(expected, actual, true)}${this.conformAssertMsg(optionalMsg)}`); 
		}		
	}
	
	clear() {
		this.subTests = null;	//lazily initialized
		//this.count = 0;
		//this.failureCount = 0;
		//this.exceptionCount = 0;
		this.currentSubTest = null;		
		this.log = '';		
	}
		
	conformAssertMsg(aMsg) {
		return aMsg ? ` -- ${aMsg}` : '';
	}
	
	formatComparedValue(value) {
		if (typeof value === 'string')
			return '"' + value + '"';
		return value;
	}
	
	conformCompare(expected, actual, shouldCompare) {
		return shouldCompare 
			? ` -- actual <${this.formatComparedValue(actual)}> should be equal to expected <${this.formatComparedValue(expected)}>`
			: '';
	}
	
	fail(msg) {
		//Record test as failed
		this.show(`Failed -- ${msg}`);
		this.assertTrue(false);
	}		
	
	static header(header) {
		this.show(`\nStarting test: ${header}`);
	}		
	
	assertEqualsSafely(a, b)  {
		//If only one is nil, then false, use XOR (no xor in js)*/
		if (a === null && b !== null) return false;
		if (a !== null && b === null) return false;		
		//We now know that if either is null, they both are
		if (a === null)
		return true;
		//Finally a safe equals
		//ASSUME -- this is prims only
		if (Array.isArray(a))
			return this.basicCompareArrays(a, b);
		if (!isNaN(a) && !isNaN(b))
			return Math.abs(a - b) <= Math.abs(0.0000000001);
		return a === b;
	}	
	
	basicCompareArrays(a, b) {
		return a.filter((ea, index) => ea !== b[index]).length === 0;
	}
	
	printSubTestSummary(testeeType) {	
		this.show(`----------- TEST SUMMARY -----------\n`);	
		for (let eachSubTest of this.getSubTests())
			this.show(eachSubTest.toString());
		this.show('');
	}
	
	printSummary(testeeType) {
		cr();
		this.printSubTestSummary(testeeType);
		let tests, count, passedCount, percent;
		tests = this.getSubTests();
		count = tests.reduce((runningCount, test) => runningCount + test.getTotalCount(), 0);
		passedCount = tests.reduce((runningCount, test) => runningCount + test.getPassedCount(), 0);
		percent = (count > 0) ? 100.0 * passedCount / count : 0;
		this.show(`----------- TEST CASE SUMMARY -----------\n`);
		this.show(`Unit Test: ${this.constructor.name}`);				
		this.show(`Unit Type: ${testeeType}`);		
		this.show(`Test Count: ${count}`);
		this.show(`Passing Count: ${passedCount}`);		
		this.show(`Passing Percent: ${percent.toFixed(0)}%`);
	}		
	
	OLDprintSummaryOLD(testeeType) {
		cr();
		let n, f, p, e, percent;
		n = this.count;
		f = this.failureCount;
		e = this.exceptionCount;
		p = n - f - e;
		if (n == 0) return;
		percent = 100.0 * p / n;
		this.show(`----------- TEST CASE SUMMARY -----------\n`);
		this.show(`Unit Test: ${this.constructor.name}`);				
		this.show(`Unit Type: ${testeeType}`);		
		this.show(`Test Count: ${n}`);
		this.show(`Passing Count: ${p}`);		
		this.show(`Failure Count: ${f}`);
		this.show(`Exception Count: ${e}`);		
		this.show(`Passing Percent: ${percent.toFixed(0)}%`);
	}	
	
	newSubTest(subTestLabel, testFunction) {
		return new Test(this, subTestLabel, testFunction); 
	}	
	
	incrementCount() {
		this.count++;
	}
	
	incrementFailureCount() {
		this.failureCount++;
	}	
	
	incrementExceptionCount() {
		//this.exceptionCount++;
		this.currentSubTest.incrementExceptions();
		//Return this for method cascading
		return this;
	}	
	
	getLog() {
		return this.log;
	}
	
	show(msg) {
		//to console
		prn(msg);
		//to local log variable
		let log = this.getLog();
		if (log.length > 0)
		log += '\n';
		log += msg;
		this.log = log;
		//Return this for method cascading
		return this;		
	}
		
	//----------------- Virtual Optional -----------------
	
	beforeEach() {
		//Called before each sub test
		//Optional override
	}
	
	afterEach() {
		//Called after each sub test
		//Optional override
	}	
	
	beforeAll() {
		//Called before any sub test has been run
		//Optional override
		this.show(`----------- TEST LOG -----------\n`);
	}
	
	afterAll() {
		//Called after any sub test has been run
		//Optional override
	}	
	
	getUnitName() {
		//virtual optional
		//Default is to extract model name as:
		//RectangleTest => Rectangle
		implementedBySubclass(this, "getUnitName");
		/*const testCaseName = this.constructor.name;
		const i = testCaseName.indexOf('Test');
		if (i <= 0)
			return 'Test Case';
		return testCaseName.substring(0, i);*/
	}
	
	//----------------- Non Virtual -----------------	
	
	getSubTests() {
		if (this.subTests === null) this.subTests = this.buildSubTests();
		return this.subTests;
	}
			
	buildSubTests() {
		//Build the sub-tests
		let properties = Object.getOwnPropertyNames(Object.getPrototypeOf(this));
		properties = properties.filter(eachProp => typeof this[eachProp] === 'function');
		properties = properties.filter(eachProp => eachProp.startsWith('test'));
		const tests = [];
		for (let p of properties) {
			tests.push(this.newSubTest(p, () => this[p]()));
		}
		return 	tests;		
	}	
	
	preTest() {
		//Check to make sure we can continue with tests
		//Safety check as they might say in airline world
		//Make sure we can find the unit model class (to test)
		let uname = null;
		try {
			uname = this.getUnitName();
		} catch(ex) {
			return ex.message;
		}
		const unitClass = coerceClass(uname);		
		if (typeof unitClass !== 'function')
			return `Unit model class "${uname}" does not exist`;
		return null;
	}	
	
}

//==============================================

//----------------------------------------------------	
//Helpers

function prn(anObject) {
	let safeValue = anObject;
	if (safeValue === undefined) safeValue = '*undefined*';
	if (safeValue === null) safeValue = '*null*';	
	console.log(safeValue.toString());
}

function separator() {
	return "--------------------------------------------";
}	

function minorSeparator() {
	return "-----------";
}	

function cr() {
	//carriage return new line
	prn("");
}	

//----------------------------------------------
//Test Case Management

let 
	testClassClasses = [],
	currentTestCaseClass = null;

function assureCurrentTestCase() {
	if (!currentTestCaseClass)
		currentTestCaseClass = testClassClasses[0];
}

function addTestCase(testCaseClassName) {
	//const testCaseClass = coerceClass(testCaseClassName);
	testClassClasses.push(testCaseClassName);
}

function newTestCase() {
	assureCurrentTestCase();
	if (currentTestCaseClass) {
		const cl = coerceClass(currentTestCaseClass);
		return new cl();
	}
	return null;
}

function setCurrentTestCase(aTestCaseClassName) {
	currentTestCaseClass = aTestCaseClassName;
}








